
#include "globals.h"
#include "externals.h"
#include "utils.h"

#include "mcc_generated_files/mcc.h"

#include "chan20.h"

/*
 K2000
 In 20 channel mode we receive 6 SPI bytes in each relay 'command'
 These are shown here in order CMD[0] CMD[1] CMD[2] CMD[3] CMD[4] CMD[5]
 
              CMD[0..5]        CMD[0..5]            CMD[0..5]        CMD[0..5]            CMD[0..5]        CMD[0..5]
 STARTUP     
 OPEN ALL     055555000000 3ms 000000000000  200us  010000555555 3ms 000000000000
 CLOSE CH1 2W 055555000000 3ms 000000000000  200us  000000555555 3ms 000000000000  200us  010000200000 3ms 0000000000000
 CLOSE CH2 2W 055555000000     000000000000         000000555555     000000000000         010000800000     0000000000000
 CLOSE CH3 2W 055555000000     000000000000         000000555555     000000000000         010002000000     0000000000000
  
 CLOSE CH1 4W 055555000000 3ms 000000000000  200us  000000555555 3ms 000000000000  200us  020000200002 3ms 0000000000000
*/

#define CMD0_NO_CONNECT_BITS 0b11111100 //The used upper 6 bits tend to be 000000 or 000001

#define CMD0_CH_OFF_BITS 0b00000001
#define CMD1_CH_OFF_BITS 0b01010101
#define CMD2_CH_OFF_BITS 0b01010101
#define CMD3_CH_OFF_BITS 0b01010101
#define CMD4_CH_OFF_BITS 0b01010101
#define CMD5_CH_OFF_BITS 0b01010101

#define CMD0_4W_ON_BIT   0b00000010
#define CMD0_4W_OFF_BIT  0b00000001

#define CMD1_CH07_ON_BIT 0b00000010
#define CMD1_CH08_ON_BIT 0b00001000
#define CMD1_CH09_ON_BIT 0b00100000
#define CMD1_CH10_ON_BIT 0b10000000
#define CMD1_ON_BITS (CMD1_CH07_ON_BIT | CMD1_CH08_ON_BIT | CMD1_CH09_ON_BIT | CMD1_CH10_ON_BIT)

#define CMD2_CH03_ON_BIT 0b00000010
#define CMD2_CH04_ON_BIT 0b00001000
#define CMD2_CH05_ON_BIT 0b00100000
#define CMD2_CH06_ON_BIT 0b10000000
#define CMD2_ON_BITS (CMD2_CH03_ON_BIT | CMD2_CH04_ON_BIT | CMD2_CH05_ON_BIT | CMD2_CH06_ON_BIT)

#define CMD3_CH19_ON_BIT 0b00000010
#define CMD3_CH20_ON_BIT 0b00001000
#define CMD3_CH01_ON_BIT 0b00100000
#define CMD3_CH02_ON_BIT 0b10000000
#define CMD3_ON_BITS (CMD3_CH19_ON_BIT | CMD3_CH20_ON_BIT | CMD3_CH01_ON_BIT | CMD3_CH02_ON_BIT)

#define CMD4_CH15_ON_BIT 0b00000010
#define CMD4_CH16_ON_BIT 0b00001000
#define CMD4_CH17_ON_BIT 0b00100000
#define CMD4_CH18_ON_BIT 0b10000000
#define CMD4_ON_BITS (CMD4_CH15_ON_BIT | CMD4_CH16_ON_BIT | CMD4_CH17_ON_BIT | CMD4_CH18_ON_BIT)

#define CMD5_CH11_ON_BIT 0b00000010
#define CMD5_CH12_ON_BIT 0b00001000
#define CMD5_CH13_ON_BIT 0b00100000
#define CMD5_CH14_ON_BIT 0b10000000
#define CMD5_ON_BITS (CMD5_CH11_ON_BIT | CMD5_CH12_ON_BIT | CMD5_CH13_ON_BIT | CMD5_CH14_ON_BIT)


static bool enable_2W_deferred = false;


void chan20handler(void)
{
    unsigned char cmd_idx;
    unsigned char cmd[6];

    if(ringbuf_cmd_wr_idx == ringbuf_cmd_rd_idx) {
        return;
    }

    //New SPI command received
    cmd_idx = ringbuf_cmd_idx[ringbuf_cmd_rd_idx];
    RINGBUF_IDX_INCR(ringbuf_cmd_rd_idx);
    //Copy command bytes to local variables
    cmd[0] = ringbuf_spi[cmd_idx];
    RINGBUF_IDX_INCR(cmd_idx);
    cmd[1] = ringbuf_spi[cmd_idx];
    RINGBUF_IDX_INCR(cmd_idx);
    cmd[2] = ringbuf_spi[cmd_idx];
    RINGBUF_IDX_INCR(cmd_idx);
    cmd[3] = ringbuf_spi[cmd_idx];
    RINGBUF_IDX_INCR(cmd_idx);
    cmd[4] = ringbuf_spi[cmd_idx];
    RINGBUF_IDX_INCR(cmd_idx);
    cmd[5] = ringbuf_spi[cmd_idx];

//    EUSART_Write(cmd[0]);
//    EUSART_Write(cmd[1]);
//    EUSART_Write(cmd[2]);
//    EUSART_Write(cmd[3]);
//    EUSART_Write(cmd[4]);
//    EUSART_Write(cmd[5]);
//    return;
    
    //Decode relay control bits cmd[6] received through SPI            

    //The 6 unused (nc) bits in cmd[0] seem to be 0b000000 or 0b000001
    //If not, something is wrong, SPI may be out of sync
    //Ignore the command and re-sync SPI in STROBE ISR
    
    if( ((cmd[0] & CMD0_NO_CONNECT_BITS) != 0) && ((cmd[0] & CMD0_NO_CONNECT_BITS) != 0x04) ) {
        //Disable SPI and re-enable in STROBE interrupt
        PIE3bits.SSP1IE = 0;
        SSP1CON1bits.SSPEN = 0;
        spi_resync = true;
        return;
    }
    
    //Mask off the unused (nc) bits
    cmd[0] &= ~CMD0_NO_CONNECT_BITS;

    if(cmd[0]==0 && cmd[1]==0 && cmd[2]==0 && cmd[3]==0 && cmd[4]==0 && cmd[5]==0) {
        //Coil power off, no action required here
    }
    else if((cmd[1] & CMD1_CH_OFF_BITS)!=0 ||
            (cmd[2] & CMD2_CH_OFF_BITS)!=0 ||
            (cmd[3] & CMD3_CH_OFF_BITS)!=0 ||
            (cmd[4] & CMD4_CH_OFF_BITS)!=0 ||
            (cmd[5] & CMD5_CH_OFF_BITS)!=0)
    {
        // ============== Switching relays off (OPEN) ==============
        relays_open_all_20_channels();
        
        //NOTE Switching all channel relays off like this generally works fine
        // except in the DMM6500 when switching between 4W and 2W measurement on
        // the SAME channel.
        // Fix is to switch to another channel, then back again.
        // Works fine on K2000 because that switches all relays off first anyway.
        
        if((cmd[0] & CMD0_4W_ON_BIT)!=0) {
            //Enable 4W, which is safe as it disconnects the two relay banks
            SENSE_ENn_LAT = 0;    //Connect sense output to DMM
            TWO_WIRE_ENn_LAT = 1; //Isolate BANK1 and BANK2
            enable_2W_deferred = false;
        }
        else if((cmd[0] & CMD0_4W_OFF_BIT)!=0) {
            //Enable 2W
            //The host DMM sometimes enables 2W here but I think
            //that is unwise, as it connects BANK1 and BANK2 while other
            //channel relays may be (momentarily) closed, about to open.
            //Defer 2W until later, when closing channel relays.
            enable_2W_deferred = true;
        }
    }
    else if((cmd[1] & CMD1_ON_BITS)!=0 || (cmd[2] & CMD2_ON_BITS)!=0 || (cmd[3] & CMD3_ON_BITS)!=0 ||
                (cmd[4] & CMD4_ON_BITS)!=0 || (cmd[5] & CMD5_ON_BITS)!=0 ||
                (cmd[0] & CMD0_4W_OFF_BIT)!=0 || enable_2W_deferred ) {
        // ============== Switching relays on (CLOSE) ==============
        //Golden rules:
        // In 2W mode never activate more than one channel
        // In 4W mode never activate more than one channel in each BANK

        //All channel relays must be open before we close any
        if(relays_are_all_channels_off()) {

            if((cmd[0] & CMD0_4W_ON_BIT)!=0) {
                //Switch to 4W (isolate the two banks)
                TWO_WIRE_ENn_LAT=1;
                SENSE_ENn_LAT = 0;    //Connect sense output to DMM
                enable_2W_deferred = false;
            }
            else if((cmd[0] & CMD0_4W_OFF_BIT)!=0  || enable_2W_deferred) {
                //Switch to 2W (connect the two banks)
                SENSE_ENn_LAT = 1;    //Disconnect sense output to DMM
                TWO_WIRE_ENn_LAT=0;
                enable_2W_deferred = false;
            }

            if((cmd[1] & CMD1_ON_BITS)!=0) {
                //CH7..CH10
                if((cmd[1] & CMD1_CH07_ON_BIT)!=0) { //CH7
                    BANK1_ODD_SetHigh();
                    CHAN7_8_SetLow();
                    if(TWO_WIRE_ENn_LAT==1) { //4-Wire Mode
                        //CH17
                        BANK2_ODD_SetHigh();
                        CHAN17_18_SetLow();
                    }
                }
                else if((cmd[1] & CMD1_CH08_ON_BIT)!=0) { //CH8
                    BANK1_EVEN_SetHigh();
                    CHAN7_8_SetLow();
                    if(TWO_WIRE_ENn_LAT==1) { //4-Wire Mode
                        //CH18
                        BANK2_EVEN_SetHigh();
                        CHAN17_18_SetLow();
                    }
                }
                else if((cmd[1] & CMD1_CH09_ON_BIT)!=0) { //CH9
                    BANK1_ODD_SetHigh();
                    CHAN9_10_SetLow();
                    if(TWO_WIRE_ENn_LAT==1) { //4-Wire Mode
                        //CH19
                        BANK2_ODD_SetHigh();
                        CHAN19_20_SetLow();
                    }
                }
                else if((cmd[1] & CMD1_CH10_ON_BIT)!=0) { //CH10
                    BANK1_EVEN_SetHigh();
                    CHAN9_10_SetLow();
                    if(TWO_WIRE_ENn_LAT==1) { //4-Wire Mode
                        //CH20
                        BANK2_EVEN_SetHigh();
                        CHAN19_20_SetLow();
                    }
                }
            }
            else if((cmd[2] & CMD2_ON_BITS)!=0) {
                //CH3..CH6
                if((cmd[2] & CMD2_CH03_ON_BIT)!=0) { //CH3
                    BANK1_ODD_SetHigh();
                    CHAN3_4_SetLow();
                    if(TWO_WIRE_ENn_LAT==1) { //4-Wire Mode
                        //CH13
                        BANK2_ODD_SetHigh();
                        CHAN13_14_SetLow();
                    }
                }
                else if((cmd[2] & CMD2_CH04_ON_BIT)!=0) { //CH4
                    BANK1_EVEN_SetHigh();
                    CHAN3_4_SetLow();
                    if(TWO_WIRE_ENn_LAT==1) { //4-Wire Mode
                        //CH14
                        BANK2_EVEN_SetHigh();
                        CHAN13_14_SetLow();
                    }
                }
                else if((cmd[2] & CMD2_CH05_ON_BIT)!=0) { //CH5
                    BANK1_ODD_SetHigh();
                    CHAN5_6_SetLow();
                    if(TWO_WIRE_ENn_LAT==1) { //4-Wire Mode
                        //CH15
                        BANK2_ODD_SetHigh();
                        CHAN15_16_SetLow();
                    }
                }
                else if((cmd[2] & CMD2_CH06_ON_BIT)!=0) { //CH6
                    BANK1_EVEN_SetHigh();
                    CHAN5_6_SetLow();
                    if(TWO_WIRE_ENn_LAT==1) { //4-Wire Mode
                        //CH16
                        BANK2_EVEN_SetHigh();
                        CHAN15_16_SetLow();
                    }
                }
            }
            else if((cmd[3] & CMD3_ON_BITS)!=0) {
                //CH19 CH20 CH1 CH2
                if((cmd[3] & CMD3_CH01_ON_BIT)!=0) { //CH1
                    BANK1_ODD_SetHigh();
                    CHAN1_2_SetLow();
                    if(TWO_WIRE_ENn_LAT==1) { //4-Wire Mode
                        //CH11
                        BANK2_ODD_SetHigh();
                        CHAN11_12_SetLow();
                    }
                }
                else if((cmd[3] & CMD3_CH02_ON_BIT)!=0) { //CH2
                    BANK1_EVEN_SetHigh();
                    CHAN1_2_SetLow();
                    if(TWO_WIRE_ENn_LAT==1) { //4-Wire Mode
                        //CH12
                        BANK2_EVEN_SetHigh();
                        CHAN11_12_SetLow();
                    }
                }
                else if((cmd[3] & CMD3_CH19_ON_BIT)!=0) { //CH19
                    BANK2_ODD_SetHigh();
                    CHAN19_20_SetLow();
                }
                else if((cmd[3] & CMD3_CH20_ON_BIT)!=0) { //CH20
                    BANK2_EVEN_SetHigh();
                    CHAN19_20_SetLow();
                }
            }
            else if((cmd[4] & CMD4_ON_BITS)!=0) {
                //CH15..CH18
                if((cmd[4] & CMD4_CH15_ON_BIT)!=0) { //CH15
                    BANK2_ODD_SetHigh();
                    CHAN15_16_SetLow();
                }
                else if((cmd[4] & CMD4_CH16_ON_BIT)!=0) { //CH16
                    BANK2_EVEN_SetHigh();
                    CHAN15_16_SetLow();
                }
                else if((cmd[4] & CMD4_CH17_ON_BIT)!=0) { //CH17
                    BANK2_ODD_SetHigh();
                    CHAN17_18_SetLow();
                }
                else if((cmd[4] & CMD4_CH18_ON_BIT)!=0) { //CH18
                    BANK2_EVEN_SetHigh();
                    CHAN17_18_SetLow();
                }
            }
            else if((cmd[5] & CMD5_ON_BITS)!=0) {
                //CH11..CH14
                if((cmd[5] & CMD5_CH11_ON_BIT)!=0) { //CH11
                    BANK2_ODD_SetHigh();
                    CHAN11_12_SetLow();
                }
                else if((cmd[5] & CMD5_CH12_ON_BIT)!=0) { //CH12
                    BANK2_EVEN_SetHigh();
                    CHAN11_12_SetLow();
                }
                else if((cmd[5] & CMD5_CH13_ON_BIT)!=0) { //CH13
                    BANK2_ODD_SetHigh();
                    CHAN13_14_SetLow();
                }
                else if((cmd[5] & CMD5_CH14_ON_BIT)!=0) { //CH14
                    BANK2_EVEN_SetHigh();
                    CHAN13_14_SetLow();
                }
            }
        }
    }
}