#include "mcc_generated_files/mcc.h"

#include "stdio.h"

#include "globals.h"
#include "utils.h"
#include "chan10.h"
#include "chan20.h"

void spiISR(void);
void strobeISR(void);
void tickISR(void);

volatile unsigned char ringbuf_spi[RINGBUF_SIZE]; //Ringbuffer to hold incoming SPI bytes
volatile unsigned char ringbuf_spi_wr_idx=0;
volatile unsigned char ringbuf_cmd_idx[RINGBUF_SIZE]; //Ringbuffer to hold index of command start
volatile unsigned char ringbuf_cmd_wr_idx=0;
volatile unsigned char ringbuf_cmd_rd_idx=0;
volatile unsigned char spi_rx_byte_cnt=0;
volatile unsigned char tick_counter=0; //decremented in timer0 interrupt
volatile unsigned char cmd[6];
volatile bool spi_resync = false;
volatile bool mode_10_channels = false;


void main(void)
{
    bool spi_clk_idle_state = false;
    unsigned char uc;

    SYSTEM_Initialize();

    relays_open_all_20_channels();
    SENSE_ENn_LAT = 1;    //Disconnect sense output to DMM

    // When using interrupts, you need to set the Global and Peripheral Interrupt Enable bits
    // Use the following macros to:
    //INTERRUPT_GlobalInterruptEnable();
    //INTERRUPT_PeripheralInterruptEnable();
    //INTERRUPT_GlobalInterruptDisable();
    //INTERRUPT_PeripheralInterruptDisable();
    
    TMR0_SetInterruptHandler(tickISR);
    
    INTERRUPT_PeripheralInterruptEnable();
    INTERRUPT_GlobalInterruptEnable();
    
    //The system RESET stays low after power on
    //In K2000 it is low for    250ms
    //In DMM6500 it is low for 3500ms
    //After that, RESET goes to 5V
    //In K2000 the SPI CLK line goes high at same time as RESET (within 2 ms anyway)
    // and idles high during operation.
    //In DMM6500 the SPI CLK line is low during RESET and idles low during operation.
    
    //Wait for system RESET to go low (if not already low), 50ms timeout
    tick_counter=50;
    while(DMM_RESET_RD6_PORT == 1 && tick_counter != 0);
    
    //Wait for system RESET end (go high) with 5000ms timeout
    for(uc=0;uc<20;uc++) {
        tick_counter=250;
        while(DMM_RESET_RD6_PORT == 0 && tick_counter != 0);
    }
    
    //Get here just after system reset, wait 10ms
    tick_counter=10;
    while(tick_counter != 0);
    
    //The SPI clock line from K2000 idles high, whilst from DMM6500 it idles low.
    //Determine idle state of SPI clock by waiting for 50 samples of the same value
    //Both use 3us wide clock pulses
    spi_clk_idle_state = IO_RC3_PORT;
    uc=50;
    do {
        if(spi_clk_idle_state != IO_RC3_PORT) {
            spi_clk_idle_state = !spi_clk_idle_state;
            uc=50;
        }
        NOP();NOP();NOP();NOP();
    }
    while(--uc != 0);
    
    //spi_clk_idle_state should now be the idle state of clock line
    
    INTERRUPT_PeripheralInterruptDisable();
    INTERRUPT_GlobalInterruptDisable();
    
    //The SPI data input must be sampled on the rising edge of clock.
    //So, for clock idle low, CKP=0, CKE=0
    //For clock idle high, CKP=1, CKE=1
    //SPI should be enabled when the clock line is in the idle state
    //PIC instruction clock Tcy must be less than SPI clock min of 3us pulses.
    //So PIC clock of 8Mhz/4=2Mhz => Tcy = 0.5uS will be fine
    
    if(spi_clk_idle_state) {
        //Clock idle high, sample on rising edge:
        SSP1CON1bits.CKP=1;
        SSP1STATbits.CKE=0;
    }
    else {    
      //Clock idle low, sample on rising edge:
      SSP1CON1bits.CKP=0;
      SSP1STATbits.CKE=1;
    }
    
    SPI1_SetInterruptHandler(spiISR);
    SSP1CON1bits.SSPEN = 1;
    PIR3bits.SSP1IF = 0;
    
    IOCCF5_SetInterruptHandler(strobeISR);
    
    //Sample the MODE input to select 10 or 20 channel mode
    mode_10_channels = DMM_MODE_10_20_PORT;

    PIE0bits.TMR0IE = 0; //Don't need the ms tick interrupt within main loop
    PIR0bits.TMR0IF = 0;
    
    INTERRUPT_PeripheralInterruptEnable();
    INTERRUPT_GlobalInterruptEnable();

    while (1)
    {
        //Monitor system RESET line
        if(DMM_RESET_PORT == 0) {
            NOP();
            NOP();
            if(DMM_RESET_PORT == 0) {
                relays_open_all_20_channels();
                SENSE_ENn_LAT = 1;    //Disconnect sense output to DMM
                TWO_WIRE_ENn_LAT = 1; //Isolate BANK1 and BANK2
            }
        }
        else {
            
            //The SPI receive interrupt needs serviced every 80us when
            //receiving relay command (3 bytes or 6 bytes) from DMM.
            //PIC16F18875 has no interrupt priority or nesting, so any other
            //enabled interrupt must be kept short and fast.
            //e.g. interrupt driven uart has caused some issues (sometimes missed SPI byte)
            //MCC interrupt handling code isn't optimal either.

            if(mode_10_channels) {
                chan10handler();
            }
            else {
                chan20handler();
            }
        }
    }
}


void tickISR(void) {
    //1ms interrupt from timer0
    if (tick_counter != 0) {
        tick_counter--;
    }    
}


void spiISR(void){
    //Save SPI bytes into ringbuffer
    ringbuf_spi[ringbuf_spi_wr_idx] = SPI1_ReadByte();
    RINGBUF_IDX_INCR(ringbuf_spi_wr_idx);
    spi_rx_byte_cnt++;
    PIR3bits.SSP1IF = 0;
}


void strobeISR(void){
    //Since there is no Slave Select signal, there is the risk of SPI bit shift
    //problem, e.g. if we have a clock line noise glitch.
    //If a problem is detected, re-sync to the strobe pulses (clear SPI bit counter)
    if(spi_resync) {
        
        //TODO test this bit shift recovery
        
        //SPI signals are in idle state during/just after STROBE pulse
        PIE3bits.SSP1IE = 0;
        SSP1CON1bits.SSPEN = 0;
        NOP();
        NOP();
        PIR3bits.SSP1IF = 0;
        PIE3bits.SSP1IE = 1;
        SSP1CON1bits.SSPEN = 1;
        spi_rx_byte_cnt=0;
        ringbuf_spi_wr_idx = 0;
        ringbuf_cmd_wr_idx = 0;
        ringbuf_cmd_rd_idx = 0; //Flush the command buffer
        spi_resync = false;
        return;
    }
    
    //Strobe pulse occurs just 4 usecs or so after last SPI clock pulse
    //So SPI service routine may not have been serviced yet
    //depending on structure of INTERRUPT_InterruptManager() 
    if(PIE3bits.SSP1IE == 1 && PIR3bits.SSP1IF == 1) {
        spiISR();
    }
    
    //Valid data should have 3 (or 6) spi bytes received between strobe pulses    
    if( (mode_10_channels == true  && spi_rx_byte_cnt == 3) ||
        (mode_10_channels == false && spi_rx_byte_cnt == 6) ) {
            ringbuf_cmd_idx[ringbuf_cmd_wr_idx] =
                (ringbuf_spi_wr_idx - spi_rx_byte_cnt)&(RINGBUF_SIZE-1);
            RINGBUF_IDX_INCR(ringbuf_cmd_wr_idx);
    }
    spi_rx_byte_cnt=0;
}


/**
 End of File
*/